import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params
import json
from pl_bolts.models.self_supervised import SwAV
from tqdm import tqdm
import numpy as np
from omegaconf import OmegaConf
from pl_bolts.models.self_supervised.swav.swav_resnet import resnet18, resnet50
import copy
from torchvision.models.utils import load_state_dict_from_url
from torchvision.models.resnet import model_urls


def sinkhorn(Q, nmb_iters):
    with torch.no_grad():
        sum_Q = torch.sum(Q)
        Q /= sum_Q

        K, B = Q.shape

        u = torch.zeros(K).cuda()
        r = torch.ones(K).cuda() / K
        c = torch.ones(B).cuda() / B

        for _ in range(nmb_iters):
            u = torch.sum(Q, dim=1)

            Q *= (r / u).unsqueeze(1)
            Q *= (c / torch.sum(Q, dim=0)).unsqueeze(0)

        return (Q / torch.sum(Q, dim=0, keepdim=True)).t().float()


def evaluation(models, config, dl, from_training, **kwargs):
    # No ensembling for this evaluation
    model = models[0]

    losses = []
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        out = model(**batch, from_training=from_training)
        losses.append(out['loss'].mean().cpu().data.numpy())

        if not from_training:
            visuals.append(out['visual'].cpu())
        # break

    if from_training:
        return {'loss': np.ndarray.mean(np.array(losses))}

    return {'loss': np.ndarray.mean(np.array(losses)),
            'visual': torch.cat(visuals)
            }


class SwaV(nn.Module):
    def __init__(self, config, **kwargs):
        super().__init__()

        self.config = copy.deepcopy(config)

        # Visual Encoder
        self.sinkhorn_iterations = config.pop('sinkhorn_iterations')
        self.hidden_mlp = config.pop('hidden_mlp')
        self.output_dim = config.pop('output_dim')
        self.nmb_prototypes = config.pop('nmb_prototypes')
        self.first_conv = config.pop('first_conv')
        self.maxpool1 = config.pop('maxpool1')

        self.crops_for_assign = config.pop('crops_for_assign')
        self.nmb_crops = config.pop('nmb_crops')
        self.temperature = config.pop('temperature')
        self.epsilon = config.pop('epsilon')

        arch = config.arch
        self.model = eval(arch)(
            normalize=True,
            hidden_mlp=self.hidden_mlp,
            output_dim=self.output_dim,
            nmb_prototypes=self.nmb_prototypes,
            first_conv=self.first_conv,
            maxpool1=self.maxpool1,
        )

        self.model.load_state_dict(state_dict=load_state_dict_from_url(model_urls[arch], progress=True), strict=False)
        self.softmax = nn.Softmax(dim=1)

        # Evaluation
        self.eval_func = evaluation

    def forward(self, images, from_training=True, **kwargs):
        if not from_training:
            return {"loss": torch.tensor(0.),
                    'visual': self.model.projection_head(self.model.forward_backbone(images.cuda()))}

        # https://github.com/PyTorchLightning/lightning-bolts/blob/master/pl_bolts/models/self_supervised/swav/swav_module.py
        # 1. normalize the prototypes
        images = images[:-1]
        with torch.no_grad():
            w = self.model.prototypes.weight.data.clone()
            w = nn.functional.normalize(w, dim=1, p=2)
            self.model.prototypes.weight.copy_(w)

        # 2. multi-res forward passes
        embedding, output = self.model(images)
        embedding = embedding.detach()
        bs = images[0].size(0)

        # 3. swav loss computation
        loss = 0
        for i, crop_id in enumerate(self.crops_for_assign):
            with torch.no_grad():
                out = output[bs * crop_id: bs * (crop_id + 1)]

                # 4. get assignments
                q = torch.exp(out / self.epsilon).t()
                q = sinkhorn(q, self.sinkhorn_iterations)[-bs:]

            # cluster assignment prediction
            subloss = 0
            for v in np.delete(np.arange(np.sum(self.nmb_crops)), crop_id):
                p = self.softmax(output[bs * v: bs * (v + 1)] / self.temperature)
                subloss -= torch.mean(torch.sum(q * torch.log(p), dim=1))
            loss += subloss / (np.sum(self.nmb_crops) - 1)
        loss /= len(self.crops_for_assign)

        return {"loss": loss}

    def __repr__(self):
        s = "SwaV\n"
        s += str(json.dumps(OmegaConf.to_container(self.config), indent=4)) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import json
import faiss
import copy

from tqdm import tqdm
from omegaconf import OmegaConf
import numpy as np
import torch.nn as nn
from vilmedic.models.utils import get_n_params
from vilmedic.blocks.vision.selfsup import MoCo


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


def run_kmeans(x,
               num_cluster,
               temperature,
               ):
    """
    Args:
        x: data to be clustered
    """

    print('performing kmeans clustering')
    results = {'im2cluster': [], 'centroids': [], 'density': []}

    for seed, num_cluster in enumerate(num_cluster):
        # intialize faiss clustering parameters
        d = x.shape[1]
        k = int(num_cluster)
        clus = faiss.Clustering(d, k)
        clus.verbose = True
        clus.niter = 20
        clus.nredo = 5
        clus.seed = seed
        clus.max_points_per_centroid = 1000
        clus.min_points_per_centroid = 10

        res = faiss.StandardGpuResources()
        cfg = faiss.GpuIndexFlatConfig()
        cfg.useFloat16 = False
        cfg.device = 0
        index = faiss.GpuIndexFlatL2(res, d, cfg)

        clus.train(x, index)

        D, I = index.search(x, 1)  # for each sample, find cluster distance and assignments
        im2cluster = [int(n[0]) for n in I]

        # get cluster centroids
        centroids = faiss.vector_to_array(clus.centroids).reshape(k, d)

        # sample-to-centroid distances for each cluster
        Dcluster = [[] for c in range(k)]
        for im, i in enumerate(im2cluster):
            Dcluster[i].append(D[im][0])

        # concentration estimation (phi)
        density = np.zeros(k)
        for i, dist in enumerate(Dcluster):
            if len(dist) > 1:
                d = (np.asarray(dist) ** 0.5).mean() / np.log(len(dist) + 10)
                density[i] = d

                # if cluster only has one point, use the max to estimate its concentration
        dmax = density.max()
        for i, dist in enumerate(Dcluster):
            if len(dist) <= 1:
                density[i] = dmax

        density = density.clip(np.percentile(density, 10),
                               np.percentile(density, 90))  # clamp extreme values for stability
        density = temperature * density / density.mean()  # scale the mean to temperature

        # convert to cuda Tensors for broadcast
        centroids = torch.Tensor(centroids).cuda()
        centroids = nn.functional.normalize(centroids, p=2, dim=1)

        im2cluster = torch.LongTensor(im2cluster).cuda()
        density = torch.Tensor(density).cuda()

        results['centroids'].append(centroids)
        results['density'].append(density)
        results['im2cluster'].append(im2cluster)

    return results


def evaluation(models, config, dl, **kwargs):
    # No ensembling for this evaluation
    model = models[0]
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        out = model(**batch, eval_mode=True)
        visuals.append(out['visual'].cpu())

    # Give PCL class the current evaluation features
    visuals = torch.cat(visuals)
    PCL.EVAL_FEATURES = copy.deepcopy(visuals.numpy())

    return {'loss': -1,
            'visual': visuals
            }


class PCL(nn.Module):
    EVAL_FEATURES = None

    def __init__(self, config, clustering_epoch_start, dl, **kwargs):
        super().__init__()

        self.clustering_epoch_start = clustering_epoch_start

        # Visual Encoder
        self.config = copy.deepcopy(config)
        self.model = MoCo(base_encoder=config.arch,
                          dim=config.low_dim,
                          r=config.pcl_r,
                          m=config.moco_m,
                          T=config.temperature,
                          mlp=config.mlp)
        # self.model = torch.nn.parallel.DistributedDataParallel(self.model)
        # Evaluation
        self.eval_func = evaluation

        #
        self.current_epoch = -1
        self.cluster_result = None

        # Loss
        self.criterion = nn.CrossEntropyLoss()

    def forward(self, images, index, eval_mode=False, epoch=None, **kwargs):
        if eval_mode:
            return {"visual": self.model(im_q=images.cuda(),
                                         is_eval=True)}

        images = torch.split(images.cuda(), [3, 3], dim=1)
        # New epoch, do clustering
        if epoch + 1 >= self.clustering_epoch_start and epoch != self.current_epoch:
            assert PCL.EVAL_FEATURES is not None
            self.cluster_result = run_kmeans(PCL.EVAL_FEATURES,
                                             self.config.num_cluster,
                                             self.config.temperature,
                                             )  # run kmeans clustering on master node
            self.current_epoch = epoch

        output, target, output_proto, target_proto = self.model(im_q=images[0],
                                                                im_k=images[1],
                                                                cluster_result=self.cluster_result,
                                                                index=index)
        accp, acc = None, None
        loss = self.criterion(output, target)
        if output_proto is not None:
            loss_proto = 0
            for proto_out, proto_target in zip(output_proto, target_proto):
                loss_proto += self.criterion(proto_out, proto_target)
                accp = accuracy(proto_out, proto_target)[0]

            # average loss across all sets of prototypes
            loss_proto /= len(self.config.num_cluster)
            loss += loss_proto

        acc = accuracy(output, target)[0]
        return {"loss": loss, "custom_print": "acc: {}, accp: {}".format(acc, accp)}

    def __repr__(self):
        s = "PCL\n"
        s += str(json.dumps(OmegaConf.to_container(self.config), indent=4)) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



from torch.utils.data import Dataset
from .base.TextDataset import TextDataset


class Seq2Seq(Dataset):
    def __init__(self, src, tgt, split, ckpt_dir, **kwargs):
        self.split = split
        self.src = TextDataset(**src, split=split, ckpt_dir=ckpt_dir, source="src")
        self.tgt = TextDataset(**tgt, split=split, ckpt_dir=ckpt_dir, source="tgt")

        # For decoding
        self.tgt_tokenizer = self.tgt.tokenizer
        self.tgt_tokenizer_max_len = self.tgt.tokenizer_max_len

        assert len(self.src) == len(self.tgt), (len(self.src), len(self.tgt))

    def __getitem__(self, index):
        return {**self.src.__getitem__(index), **self.tgt.__getitem__(index)}

    def get_collate_fn(self):
        def collate_fn(batch):
            tgt = self.tgt.get_collate_fn()(batch)
            tgt['decoder_input_ids'] = tgt.pop('input_ids')
            tgt['decoder_attention_mask'] = tgt.pop('attention_mask')
            collated = {**self.src.get_collate_fn()(batch), **tgt}
            return collated

        return collate_fn

    def __len__(self):
        return len(self.src)

    def __repr__(self):
        return "Seq2Seq\n" + str(self.src) + '\n' + str(self.tgt)

    def inference(self, src=None, tgt=None):
        if src is None and tgt is None:
            return dict()

        batch = {}
        if src is not None:
            batch.update(self.src.inference(src))

        if tgt is not None:
            tgt = self.tgt.inference(tgt)
            tgt['decoder_input_ids'] = tgt.pop('input_ids')
            tgt['decoder_attention_mask'] = tgt.pop('attention_mask')
            batch.update(tgt)

        assert len(set([len(v) for k, v in batch.items()])) == 1, 'elements in batch do not have the same size'
        return batch



import os
from appdirs import user_cache_dir

DATA_DIR = __file__.replace('vilmedic/constants.py', 'data')

CACHE_DIR = user_cache_dir("vilmedic")
EXTRA_CACHE_DIR = os.path.join(CACHE_DIR, "extras")
MODEL_ZOO_CACHE_DIR = os.path.join(CACHE_DIR, "zoo", "models")


from .NLG.rouge.rouge import Rouge, Rouge1, Rouge2, RougeL
from .NLG.bleu.bleu import Bleu
from .NLG.meteor.meteor import Meteor
from .NLG.ciderD.ciderD import CiderD
from .NLG.bertscore.bertscore import BertScore
from .NLG.mauve_ import MauveScorer
from .NLG.ciderD_RL.ciderD_RL import CiderDRL
from .CheXbert.chexbert import CheXbert
from .RadEntityMatchExact.RadEntityMatchExact import RadEntityMatchExact
from .RadEntityNLI.RadEntityNLI import RadEntityNLI
from .StanfordCTAbdAcc.StanfordCTAbdAcc import StanfordCTAbdAcc
